from keras.layers import Dense
from keras.models import Sequential
from keras.optimizers import SGD
from keras.initializers import normal, glorot_normal

from deepreplay.callbacks import ReplayData
from deepreplay.replay import Replay
from deepreplay.plot import compose_animations, compose_plots

from sklearn.datasets import make_moons

import matplotlib.pyplot as plt

group_name = 'moons'

X, y = make_moons(n_samples=2000, random_state=27, noise=0.03)

sgd = SGD(lr=0.01)

glorot_initializer = glorot_normal(seed=42)
normal_initializer = normal(seed=42)

replaydata = ReplayData(X, y, filename='moons_dataset.h5', group_name=group_name)

model = Sequential()
model.add(Dense(input_dim=2,
                units=4,
                kernel_initializer=glorot_initializer,
                activation='tanh'))
model.add(Dense(units=2,
                kernel_initializer=glorot_initializer,
                activation='tanh',
                name='hidden'))

model.add(Dense(units=1,
                kernel_initializer=normal_initializer,
                activation='sigmoid',
                name='output'))

model.compile(loss='binary_crossentropy',
              optimizer=sgd,
              metrics=['acc'])

model.fit(X, y, epochs=200, batch_size=16, callbacks=[replaydata])

replay = Replay(replay_filename='moons_dataset.h5', group_name=group_name)

fig = plt.figure(figsize=(12, 6))
ax_fs = plt.subplot2grid((2, 4), (0, 0), colspan=2, rowspan=2)
ax_ph_neg = plt.subplot2grid((2, 4), (0, 2))
ax_ph_pos = plt.subplot2grid((2, 4), (1, 2))
ax_lm = plt.subplot2grid((2, 4), (0, 3))
ax_lh = plt.subplot2grid((2, 4), (1, 3))

fs = replay.build_feature_space(ax_fs, layer_name='hidden',
                                xlim=(-1, 2), ylim=(-.5, 1),
                                display_grid=False, scale_fixed=False)
ph = replay.build_probability_histogram(ax_ph_neg, ax_ph_pos)
lh = replay.build_loss_histogram(ax_lh)
lm = replay.build_loss_and_metric(ax_lm, 'acc')

sample_figure = compose_plots([fs, ph, lm, lh], 80)
sample_figure.savefig('moons.png', dpi=120, format='png')

sample_anim = compose_animations([fs, ph, lm, lh])
sample_anim.save(filename='moons.mp4', dpi=120, fps=5)



from __future__ import division
import os
import numpy as np
import h5py
from keras.callbacks import Callback

class ReplayData(Callback):
    """Creates an instance of a Keras Callback to collect information
    about the training process, mostly the weights, and store them
    in a group of a HDF5 file, together with inputs and targets passed
    as arguments.

    It also saves the Keras model itself, as an independent HDF5 file,
    named after the group_name with '_model' appended to it.

    Parameters
    ----------
    inputs: ndarray
        An array with the inputs to be used during the training of the
        model. These inputs are going to be later used to generate
        statistics and plots.
    targets: ndarray
        An array with the targets to be used during the training of the
        model. These targets are going to be later used to generate
        statistics and plots.
    filename: String
        HDF5 filename to be used to store the collected information.
        It can be an existing file, as the data will be appended to it.
    group_name: String
        Group inside the HDF5 file where the information is to be
        saved. If the informed group name already exists, it will throw
        an exception.
    """
    def __init__(self, inputs, targets, filename, group_name):
        super(ReplayData, self).__init__()
        self.handler = h5py.File('{}'.format(filename), 'a')
        self.inputs = inputs
        self.targets = targets.reshape(-1, 1)
        self.filepath = os.path.split(filename)[0]
        self.filename = filename
        self.group = None
        self.group_name = group_name
        self.current_epoch = -1
        self.n_epochs = 0
        return

    def _append_weights(self):
        """Appends the weights to the corresponding layer and weight sequence inside the group, for the current epoch.
        """
        for i, layer in enumerate(self.model.layers):
            layer_weights = layer.get_weights()
            for j, weights in enumerate(layer_weights):
                self.group['layer{}'.format(i)]['weights{}'.format(j)][self.current_epoch + 1] = weights

    def on_train_begin(self, logs={}):
        self.model.save(os.path.join(self.filepath, '{}_model.h5'.format(self.group_name)))
        self.n_epochs = self.params['epochs']

        self.group = self.handler.create_group(self.group_name)
        self.group.attrs['samples'] = self.params['samples']
        self.group.attrs['batch_size'] = self.params['batch_size']
        self.group.attrs['n_batches'] = np.ceil(self.params['samples'] / self.params['batch_size']).astype(np.int)
        self.group.attrs['n_epochs'] = self.n_epochs
        self.group.attrs['n_layers'] = len(self.model.layers)
        try:
            # Python 2
            self.group.attrs['activation_functions'] = [layer.activation.func_name
                                                        if hasattr(layer, 'activation')
                                                        else ''
                                                        for layer in self.model.layers]
        except AttributeError:
            # Python 3
            self.group.attrs['activation_functions'] = [np.string_(layer.activation.__name__)
                                                        if hasattr(layer, 'activation')
                                                        else np.string_('')
                                                        for layer in self.model.layers]
        self.group.create_dataset('inputs', data=self.inputs)
        self.group.create_dataset('targets', data=self.targets)

        self.group.create_dataset('loss', shape=(self.n_epochs,), dtype='f')
        for metric in self.model.metrics:
            self.group.create_dataset(metric, shape=(self.n_epochs,), dtype='f')

        for i, layer in enumerate(self.model.layers):
            layer_grp = self.group.create_group('layer{}'.format(i))
            layer_weights = layer.get_weights()
            for j, weights in enumerate(layer_weights):
                layer_grp.create_dataset('weights{}'.format(j),
                                         shape=(self.n_epochs + 1, ) + weights.shape,
                                         dtype='f')
        self._append_weights()
        return

    def on_train_end(self, logs={}):
        self.handler.close()
        return

    def on_epoch_begin(self, epoch, logs={}):
        self.current_epoch = epoch
        return

    def on_epoch_end(self, epoch, logs={}):
        self._append_weights()
        self.group['loss'][epoch] = logs.get('loss')
        for metric in self.model.metrics:
            self.group[metric][epoch] = logs.get(metric, np.nan)
        return



from keras.layers import Dense
from keras.models import Sequential
from keras.optimizers import SGD
from keras.initializers import glorot_normal, normal

from deepreplay.datasets.parabola import load_data
from deepreplay.callbacks import ReplayData
from deepreplay.replay import Replay
from deepreplay.plot import compose_animations, compose_plots

import matplotlib.pyplot as plt

X, y = load_data()

sgd = SGD(lr=0.05)

for activation in ['sigmoid', 'tanh', 'relu']:
    glorot_initializer = glorot_normal(seed=42)
    normal_initializer = normal(seed=42)

    replaydata = ReplayData(X, y, filename='comparison_activation_functions.h5', group_name=activation)

    model = Sequential()
    model.add(Dense(input_dim=2,
                    units=2,
                    kernel_initializer=glorot_initializer,
                    activation=activation,
                    name='hidden'))

    model.add(Dense(units=1,
                    kernel_initializer=normal_initializer,
                    activation='sigmoid',
                    name='output'))

    model.compile(loss='binary_crossentropy',
                  optimizer=sgd,
                  metrics=['acc'])

    model.fit(X, y, epochs=150, batch_size=16, callbacks=[replaydata])

fig, axs = plt.subplots(1, 3, figsize=(12, 4))

replays = []
for activation in ['sigmoid', 'tanh', 'relu']:
    replays.append(Replay(replay_filename='comparison_activation_functions.h5', group_name=activation))

spaces = []
for ax, replay, activation in zip(axs, replays, ['sigmoid', 'tanh', 'relu']):
    space = replay.build_feature_space(ax, layer_name='hidden')
    space.set_title(activation)
    spaces.append(space)

sample_figure = compose_plots(spaces, 80)
sample_figure.savefig('comparison.png', dpi=120, format='png')

#sample_anim = compose_animations(spaces)
#sample_anim.save(filename='comparison.mp4', dpi=120, fps=5)




from __future__ import division
import numpy as np
import h5py
import keras.backend as K
from keras.models import load_model
from .plot import build_2d_grid, FeatureSpace, ProbabilityHistogram, LossHistogram, LossAndMetric
from .plot import FeatureSpaceData, FeatureSpaceLines, ProbHistogramData, LossHistogramData, LossAndMetricData

TRAINING_MODE = 1
TEST_MODE = 0

class Replay(object):
    """Creates an instance of Replay, to process information collected
    by the callback and generate data to feed the supported visualiza-
    tions.

    Parameters
    ----------
    replay_filename: String
        HDF5 filename used by the callback to store the training
        information.
    group_name: String
        Group inside the HDF5 file where the information was saved.
    model_filename: String, optional
        HDF5 filename of the saved Keras model. Default is the
        group_name with '_model' appended to it.

    Attributes
    ----------
    feature_space: (FeatureSpace, FeatureSpaceData)
        FeatureSpace object to be used for plotting and animating;
        namedtuple containing information about original grid lines,
        data points and predictions.

    loss_histogram: (LossHistogram, LossHistogramData)
        LossHistogram object to be used for plotting and animating;
        namedtuple containing information about example's losses.

    loss_and_metric: (LossAndMetric, LossAndMetricData)
        LossAndMetric object to be used for plotting and animating;
        namedtuple containing information about loss and a given
        metric.

    probability_histogram: (ProbabilityHistogram, ProbHistogramData)
        ProbabilityHistogram object to be used for plotting and
        animating; namedtuple containing information about
        classification probabilities and targets.

    training_loss: ndarray
        An array of shape (n_epochs, ) with training loss as reported
        by Keras at the end of each epoch.
    """
    def __init__(self, replay_filename, group_name, model_filename=''):
        # Set learning phase to TEST
        self.learning_phase = TEST_MODE

        # If not informed, defaults to '_model' suffix
        if model_filename == '':
            model_filename = '{}_model.h5'.format(group_name)

        # Loads Keras model
        self.model = load_model(model_filename)
        # Loads ReplayData file
        self.replay_data = h5py.File('{}'.format(replay_filename), 'r')
        self.group_name = group_name
        self.group = self.replay_data[self.group_name]

        # Retrieves some basic information from the replay data
        self.inputs = self.group['inputs'][:]
        self.targets = self.group['targets'][:]
        self.n_epochs = self.group.attrs['n_epochs']
        self.n_layers = self.group.attrs['n_layers']
        # Retrieves weights as a list, each element being one epoch
        self.weights = self._retrieve_weights()

        # Gets Tensors for the weights in the same order as the layers
        # Keras' model.weights returns the Tensors in a different order!
        self._model_weights = [w for layer in self.model.layers for w in layer.weights]

        ### Functions
        # Keras function to get the outputs, given inputs and weights
        self._get_output = K.function(inputs=[K.learning_phase()] + self.model.inputs + self._model_weights,
                                      outputs=[self.model.layers[-1].output])
        # Keras function to get the loss and metrics, given inputs, targets, weights and sample weights
        self._get_metrics = K.function(inputs=[K.learning_phase()] + self.model.inputs + self.model.targets +
                                              self._model_weights + self.model.sample_weights,
                                       outputs=[self.model.total_loss] + self.model.metrics_tensors)
        # Keras function to compute the binary cross entropy, given inputs, targets, weights and sample weights
        self._get_binary_crossentropy = K.function(inputs=[K.learning_phase()] + self.model.inputs +
                                                          self.model.targets + self._model_weights +
                                                          self.model.sample_weights,
                                                   outputs=[K.binary_crossentropy(self.model.targets[0],
                                                                                  self.model.outputs[0])])

        # Attributes for the visualizations - Data
        self._feature_space_data = None
        self._loss_hist_data = None
        self._loss_and_metric_data = None
        self._prob_hist_data = None
        self._decision_boundary_data = None
        # Attributes for the visualizations - Plot objects
        self._feature_space_plot = None
        self._loss_hist_plot = None
        self._loss_and_metric_plot = None
        self._prob_hist_plot = None
        self._decision_boundary_plot = None

    def _retrieve_weights(self):
        # Generates ranges for the number of different weight arrays in each layer
        n_weights = [range(len(self.group['layer{}'.format(l)]))
                     for l in range(self.n_layers)]
        # Retrieves weights for each layer and sequence of weights
        weights = [np.array(self.group['layer{}'.format(l)]['weights{}'.format(w)])
                   for l, ws in enumerate(n_weights)
                   for w in ws]
        # Since initial weights are also saved, there are n_epochs + 1 elements in total
        return [[w[epoch] for w in weights] for epoch in range(self.n_epochs + 1)]

    def _make_function(self, inputs, layer):
        """Creates a Keras function to return the output of the
        `layer` argument, given inputs and weights
        """
        return K.function(inputs=[K.learning_phase()] + inputs + self._model_weights,
                          outputs=[layer.output])

    def _predict_proba(self, inputs, weights):
        return self._get_output([self.learning_phase, inputs] + weights)

    @property
    def decision_boundary(self):
        return self._decision_boundary_plot, self._decision_boundary_data

    @property
    def feature_space(self):
        return self._feature_space_plot, self._feature_space_data

    @property
    def loss_histogram(self):
        return self._loss_hist_plot, self._loss_hist_data

    @property
    def loss_and_metric(self):
        return self._loss_and_metric_plot, self._loss_and_metric_data

    @property
    def probability_histogram(self):
        return self._prob_hist_plot, self._prob_hist_data

    @property
    def training_loss(self):
        return self.group['loss'][:]

    def get_training_metric(self, metric_name):
        """Returns corresponding metric as reported by Keras at the
        end of each epoch.

        Parameters
        ----------
        metric_name: String
            Metric to return values for.

        Returns
        -------
        metric: ndarray
            An array of shape (n_epochs, ) with the metric as reported
            by Keras at the end of each epoch.
            If the metric was not computed, returns an array of zeros
            with the same shape.
        """
        try:
            metric = self.group[metric_name][:]
        except KeyError:
            metric = np.zeros(shape=(self.n_epochs,))
        return metric

    def predict_proba(self, epoch_start=0, epoch_end=-1):
        """Generates class probability predictions for the inputs
        samples by epoch.

        Parameters
        ----------
        epoch_start: int, optional
            Initial epoch to return predicted probabilities for.
        epoch_end: int, optional
            Final epoch to return predicted probabilities for.

        Returns
        -------
        probas: ndarray
            An array of shape (n_epochs, n_samples, 2) of probabi-
            lity predictions.
        """
        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)

        probas = []
        # For each epoch, uses the corresponding weights
        for epoch in range(epoch_start, epoch_end + 1):
            weights = self.weights[epoch]
            probas.append(self._predict_proba(self.inputs, weights)[0])
        probas = np.array(probas)
        return probas

    def build_loss_histogram(self, ax, epoch_start=0, epoch_end=-1):
        """Builds a LossHistogram object to be used for plotting and
        animating.
        The underlying data, that is, the binary cross-entropy loss
        per epoch and sample, can be later accessed as the second
        element of the `loss_histogram` property.

        Only binary cross entropy loss is supported!

        Parameters
        ----------
        ax: AxesSubplot
            Subplot of a Matplotlib figure.
        epoch_start: int, optional
            First epoch to consider.
        epoch_end: int, optional
            Last epoch to consider.

        Returns
        -------
        loss_hist_plot: LossHistogram
            An instance of a LossHistogram object to make plots and
            animations.
        """
        if self.model.loss != 'binary_crossentropy':
            raise NotImplementedError("Only binary cross-entropy is supported!")

        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)
        binary_xentropy = []

        # For each epoch, uses the corresponding weights
        for epoch in range(epoch_start, epoch_end + 1):
            weights = self.weights[epoch]

            # Sample weights fixed to one!
            inputs = [self.learning_phase, self.inputs, self.targets] + weights + [np.ones(shape=self.inputs.shape[0])]
            binary_xentropy.append(self._get_binary_crossentropy(inputs=inputs)[0].squeeze())

        binary_xentropy = np.array(binary_xentropy)

        self._loss_hist_data = LossHistogramData(loss=binary_xentropy)
        self._loss_hist_plot = LossHistogram(ax).load_data(self._loss_hist_data)
        return self._loss_hist_plot

    def build_loss_and_metric(self, ax, metric_name, epoch_start=0, epoch_end=-1):
        """Builds a LossAndMetric object to be used for plotting and
        animating.
        The underlying data, that is, the loss and metric per epoch,
        can be later accessed as the second element of the
        `loss_and_metric` property.

        Parameters
        ----------
        ax: AxesSubplot
            Subplot of a Matplotlib figure.
        metric_name: String
            Metric to return values for.
        epoch_start: int, optional
            First epoch to consider.
        epoch_end: int, optional
            Last epoch to consider.

        Returns
        -------
        loss_and_metric_plot: LossAndMetric
            An instance of a LossAndMetric object to make plots and
            animations.
        """
        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)

        evaluations = []
        # For each epoch, uses the corresponding weights
        for epoch in range(epoch_start, epoch_end + 1):
            weights = self.weights[epoch]

            # Sample weights fixed to one!
            inputs = [self.learning_phase, self.inputs, self.targets] + weights + [np.ones(shape=self.inputs.shape[0])]
            evaluations.append(self._get_metrics(inputs=inputs))

        evaluations = np.array(evaluations)
        loss = evaluations[:, 0]
        try:
            metric = evaluations[:, self.model.metrics_names.index(metric_name)]
        except ValueError:
            metric = np.zeros(shape=(epoch_end - epoch_start + 1,))

        self._loss_and_metric_data = LossAndMetricData(loss=loss, metric=metric, metric_name=metric_name)
        self._loss_and_metric_plot = LossAndMetric(ax).load_data(self._loss_and_metric_data)
        return self._loss_and_metric_plot

    def build_probability_histogram(self, ax_negative, ax_positive, epoch_start=0, epoch_end=-1):
        """Builds a ProbabilityHistogram object to be used for plotting
        and animating.
        The underlying data, that is, the predicted probabilities
        and corresponding targets per epoch and sample, can be
        later accessed as the second element of the
        `probability_histogram` property.

        Only binary classification is supported!

        Parameters
        ----------
        ax_negative: AxesSubplot
            Subplot of a Matplotlib figure.
        ax_positive: AxesSubplot
            Subplot of a Matplotlib figure.
        epoch_start: int, optional
            First epoch to consider.
        epoch_end: int, optional
            Last epoch to consider.

        Returns
        -------
        prob_hist_plot: ProbabilityHistogram
            An instance of a ProbabilityHistogram object to make plots
            and animations.
        """
        if self.model.loss != 'binary_crossentropy':
            raise NotImplementedError("Only binary cross-entropy is supported!")

        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)

        self._prob_hist_data = ProbHistogramData(prob=self.predict_proba(epoch_start, epoch_end), target=self.targets)
        self._prob_hist_plot = ProbabilityHistogram(ax_negative, ax_positive).load_data(self._prob_hist_data)
        return self._prob_hist_plot

    def build_decision_boundary(self, ax, contour_points=1000, xlim=(-1, 1), ylim=(-1, 1), display_grid=True,
                                epoch_start=0, epoch_end=-1):
        """Builds a FeatureSpace object to be used for plotting and
        animating the raw inputs and the decision boundary.
        The underlying data, that is, grid lines, inputs and contour
        lines, as well as the corresponding predictions for the
        contour lines, can be later accessed as the second element of
        the  `decision_boundary` property.

        Only inputs with 2 dimensions are supported!

        Parameters
        ----------
        ax: AxesSubplot
            Subplot of a Matplotlib figure.
        contour_points: int, optional
            Number of points in each axis of the contour.
            Default is 1,000.
        xlim: tuple of ints, optional
            Boundaries for the X axis of the grid.
        ylim: tuple of ints, optional
            Boundaries for the Y axis of the grid.
        display_grid: boolean, optional
            If True, display grid lines (for 2-dimensional inputs).
            Default is True.
        epoch_start: int, optional
            First epoch to consider.
        epoch_end: int, optional
            Last epoch to consider.

        Returns
        -------
        decision_boundary_plot: FeatureSpace
            An instance of a FeatureSpace object to make plots and
            animations.
        """
        input_dims = self.model.input_shape[-1]
        assert input_dims == 2, 'Only layers with 2-dimensional inputs are supported!'

        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)

        X = self.inputs
        y = self.targets

        y_ind = y.squeeze().argsort()
        X = X.squeeze()[y_ind].reshape(X.shape)
        y = y.squeeze()[y_ind]

        n_classes = len(np.unique(y))

        # Builds a 2D grid and the corresponding contour coordinates
        grid_lines = np.array([])
        if display_grid:
            grid_lines = build_2d_grid(xlim, ylim)

        contour_lines = build_2d_grid(xlim, ylim, contour_points, contour_points)
        get_predictions = self._make_function(self.model.inputs, self.model.layers[-1])

        bent_lines = []
        bent_inputs = []
        bent_contour_lines = []
        bent_preds = []
        # For each epoch, uses the corresponding weights
        for epoch in range(epoch_start, epoch_end + 1):
            weights = self.weights[epoch]

            bent_lines.append(grid_lines)
            bent_inputs.append(X)
            bent_contour_lines.append(contour_lines)

            inputs = [TEST_MODE, contour_lines.reshape(-1, 2)] + weights
            output_shape = (contour_lines.shape[:2]) + (-1,)
            # Makes predictions for each point in the contour surface
            bent_preds.append((get_predictions(inputs=inputs)[0].reshape(output_shape) > .5).astype(np.int))

        # Makes lists into ndarrays and wrap them as namedtuples
        bent_inputs = np.array(bent_inputs)
        bent_lines = np.array(bent_lines)
        bent_contour_lines = np.array(bent_contour_lines)
        bent_preds = np.array(bent_preds)

        line_data = FeatureSpaceLines(grid=grid_lines, input=X, contour=contour_lines)
        bent_line_data = FeatureSpaceLines(grid=bent_lines, input=bent_inputs, contour=bent_contour_lines)
        self._decision_boundary_data = FeatureSpaceData(line=line_data, bent_line=bent_line_data,
                                                        prediction=bent_preds, target=y)

        # Creates a FeatureSpace plot object and load data into it
        self._decision_boundary_plot =  FeatureSpace(ax, True).load_data(self._decision_boundary_data)
        return self._decision_boundary_plot

    def build_feature_space(self, ax, layer_name, contour_points=1000, xlim=(-1, 1), ylim=(-1, 1), scale_fixed=True,
                            display_grid=True, epoch_start=0, epoch_end=-1):
        """Builds a FeatureSpace object to be used for plotting and
        animating.
        The underlying data, that is, grid lines, inputs and contour
        lines, before and after the transformations, as well as the
        corresponding predictions for the contour lines, can be
        later accessed as the second element of the `feature_space`
        property.

        Only layers with 2 hidden units are supported!

        Parameters
        ----------
        ax: AxesSubplot
            Subplot of a Matplotlib figure.
        layer_name: String
            Layer to be used for building the space.
        contour_points: int, optional
            Number of points in each axis of the contour.
            Default is 1,000.
        xlim: tuple of ints, optional
            Boundaries for the X axis of the grid.
        ylim: tuple of ints, optional
            Boundaries for the Y axis of the grid.
        scaled_fixed: boolean, optional
            If True, axis scales are fixed to the maximum from beginning.
            Default is True.
        display_grid: boolean, optional
            If True, display grid lines (for 2-dimensional inputs).
            Default is True.
        epoch_start: int, optional
            First epoch to consider.
        epoch_end: int, optional
            Last epoch to consider.

        Returns
        -------
        feature_space_plot: FeatureSpace
            An instance of a FeatureSpace object to make plots and
            animations.
        """
        # Finds the layer by name,
        layer = self.model.get_layer(layer_name)
        assert layer.output_shape == (None, 2), 'Only layers with 2-dimensional outputs are supported!'

        if epoch_end == -1:
            epoch_end = self.n_epochs
        epoch_end = min(epoch_end, self.n_epochs)

        X = self.inputs
        y = self.targets
        # Generates an indexing to fully separate negative and positive classes
        # It is not quite 'unshuffling', as it does not care about the order of the examples inside each class
        y_ind = y.squeeze().argsort()
        X = X.squeeze()[y_ind].reshape(X.shape)
        y = y.squeeze()[y_ind]

        input_dims = self.model.input_shape[-1]
        n_classes = len(np.unique(y))

        # Builds a 2D grid and the corresponding contour coordinates
        grid_lines = np.array([])
        contour_lines = np.array([])
        if input_dims == 2 and display_grid:
            grid_lines = build_2d_grid(xlim, ylim)
            contour_lines = build_2d_grid(xlim, ylim, contour_points, contour_points)

        # Creates Keras functions to get activations for the specified layer
        get_activations = self._make_function(self.model.inputs, layer)
        # Creates Keras function to get outputs of the last layer
        get_predictions = self._make_function(self.model.inputs, self.model.layers[-1])
        get_pred_from_act = self._make_function([layer.output], self.model.layers[-1])

        # Initializes "bent" variables, that is, the results of the transformations
        bent_lines = []
        bent_inputs = []
        bent_contour_lines = []
        bent_preds = []

        # For each epoch, uses the corresponding weights
        for epoch in range(epoch_start, epoch_end + 1):
            weights = self.weights[epoch]

            # Transforms the inputs
            inputs = [TEST_MODE, X] + weights
            bent_inputs.append(get_activations(inputs=inputs)[0])

            if input_dims == 2 and display_grid:
                # Transforms the grid lines
                inputs = [TEST_MODE, grid_lines.reshape(-1, 2)] + weights
                output_shape = (grid_lines.shape[:2]) + (-1,)
                bent_lines.append(get_activations(inputs=inputs)[0].reshape(output_shape))

                inputs = [TEST_MODE, contour_lines.reshape(-1, 2)] + weights
                output_shape = (contour_lines.shape[:2]) + (-1,)
                bent_contour_lines.append(get_activations(inputs=inputs)[0].reshape(output_shape))
                # Makes predictions for each point in the contour surface
                bent_preds.append((get_predictions(inputs=inputs)[0].reshape(output_shape) > .5).astype(np.int))

        bent_inputs = np.array(bent_inputs)

        if (input_dims > 2) or (not display_grid):
            xlim = (bent_inputs[:, :, 0].min(), bent_inputs[:, :, 0].max())
            ylim = (bent_inputs[:, :, 1].min(), bent_inputs[:, :, 1].max())
            grid_contour_lines = build_2d_grid(xlim, ylim, contour_points, contour_points)
            # For each epoch, uses the corresponding weights
            for epoch in range(epoch_start, epoch_end + 1):
                weights = self.weights[epoch]

                if not scale_fixed:
                    xlim = (bent_inputs[epoch, :, 0].min(), bent_inputs[epoch, :, 0].max())
                    ylim = (bent_inputs[epoch, :, 1].min(), bent_inputs[epoch, :, 1].max())
                    grid_contour_lines = build_2d_grid(xlim, ylim, contour_points, contour_points)
                # Transforms the contour lines
                bent_contour_lines.append(grid_contour_lines)

                inputs = [TEST_MODE, bent_contour_lines[-1].reshape(-1, 2)] + weights
                output_shape = (bent_contour_lines[-1].shape[:2]) + (-1,)
                bent_preds.append((get_pred_from_act(inputs=inputs)[0].reshape(output_shape) > .5).astype(np.int))

        # Makes lists into ndarrays and wrap them as namedtuples
        bent_lines = np.array(bent_lines)
        bent_contour_lines = np.array(bent_contour_lines)
        bent_preds = np.array(bent_preds)

        line_data = FeatureSpaceLines(grid=grid_lines, input=X, contour=contour_lines)
        bent_line_data = FeatureSpaceLines(grid=bent_lines, input=bent_inputs, contour=bent_contour_lines)
        self._feature_space_data = FeatureSpaceData(line=line_data, bent_line=bent_line_data,
                                                    prediction=bent_preds, target=y)

        # Creates a FeatureSpace plot object and load data into it
        self._feature_space_plot =  FeatureSpace(ax, scale_fixed).load_data(self._feature_space_data)
        return self._feature_space_plot



from keras.layers import Dense, Activation, BatchNormalization
from keras.models import Sequential
from keras.optimizers import SGD
from keras.initializers import glorot_normal, normal
from deepreplay.datasets.parabola import load_data
from deepreplay.callbacks import ReplayData

X, y = load_data()

sgd = SGD(lr=0.05)

def basic_model(activation, initializers):
    model = Sequential()
    model.add(Dense(units=2,
                    input_dim=2,
                    kernel_initializer=initializers[0],
                    activation=activation,
                    name='hidden'))

    model.add(Dense(units=1,
                    kernel_initializer=initializers[1],
                    activation='sigmoid',
                    name='output'))
    return model

def bn_model(activation, initializers):
    model = Sequential()
    model.add(Dense(units=2,
                    input_dim=2,
                    kernel_initializer=initializers[0],
                    name='hidden_linear'))
    model.add(BatchNormalization(name='hidden_bn'))
    model.add(Activation(activation, name='hidden_activation'))

    model.add(Dense(units=1,
                    kernel_initializer=initializers[1],
                    activation='sigmoid',
                    name='output'))
    return model


for seed in range(100):
    print('Using seed {}')
    replay = ReplayData(X, y, filename='part2_relu.h5', group_name='seed{:03}'.format(seed))

    glorot_initializer = glorot_normal(seed=seed)
    normal_initializer = normal(seed=42)

    model = basic_model('relu', [glorot_initializer, normal_initializer])

    model.compile(loss='binary_crossentropy',
                  optimizer=sgd,
                  metrics=['acc'])

    model.fit(X, y, epochs=150, batch_size=16, callbacks=[replay])



from __future__ import division
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
from collections import namedtuple
from matplotlib import animation
matplotlib.rcParams['animation.writer'] = 'ffmpeg'
sns.set_style('white')

FeatureSpaceData = namedtuple('FeatureSpaceData', ['line', 'bent_line', 'prediction', 'target'])
FeatureSpaceLines = namedtuple('FeatureSpaceLines', ['grid', 'input', 'contour'])
LossAndMetricData = namedtuple('LossAndMetricData', ['loss', 'metric', 'metric_name'])
ProbHistogramData = namedtuple('ProbHistogramData', ['prob', 'target'])
LossHistogramData = namedtuple('LossHistogramData', ['loss'])

def build_2d_grid(xlim, ylim, n_lines=11, n_points=1000):
    """Returns a 2D grid of boundaries given by `xlim` and `ylim`,
     composed of `n_lines` evenly spaced lines of `n_points` each.

    Parameters
    ----------
    xlim : tuple of 2 ints
        Boundaries for the X axis of the grid.
    ylim : tuple of 2 ints
        Boundaries for the Y axis of the grid.
    n_lines : int, optional
        Number of grid lines. Default is 11.
        If n_lines equals n_points, the grid can be used as
        coordinates for the surface of a contourplot.
    n_points: int, optional
        Number of points in each grid line. Default is 1,000.

    Returns
    -------
    lines : ndarray
        For the cases where n_lines is less than n_points, it
        returns an array of shape (2 * n_lines, n_points, 2)
        containing both vertical and horizontal lines of the grid.
        If n_lines equals n_points, it returns an array of shape
        (n_points, n_points, 2), containing all evenly spaced
        points inside the grid boundaries.
    """
    xs = np.linspace(*xlim, num=n_lines)
    ys = np.linspace(*ylim, num=n_points)
    x0, y0 = np.meshgrid(xs, ys)
    lines_x0 = np.atleast_3d(x0.transpose())
    lines_y0 = np.atleast_3d(y0.transpose())

    xs = np.linspace(*xlim, num=n_points)
    ys = np.linspace(*ylim, num=n_lines)
    x1, y1 = np.meshgrid(xs, ys)
    lines_x1 = np.atleast_3d(x1)
    lines_y1 = np.atleast_3d(y1)

    vertical_lines = np.concatenate([lines_x0, lines_y0], axis=2)
    horizontal_lines = np.concatenate([lines_x1, lines_y1], axis=2)

    if n_lines != n_points:
        lines = np.concatenate([vertical_lines, horizontal_lines], axis=0)
    else:
        lines = vertical_lines

    return lines

def compose_animations(objects, epoch_start=0, epoch_end=-1, title=''):
    """Compose a single animation from several objects associated with
    subplots of a single figure.

    Parameters
    ----------
    objects: list of plot objects
        Plot objects returned using one of the 'build' methods of the
        Replay class. All the corresponding subplots associated with
        the objects must belong to the same figure.
    epoch_start: int, optional
        Epoch to start the animation from.
    epoch_end: int, optional
        Epoch to end the animation.
    title: String, optional
        Text to be used in the title, preceding the epoch information.

    Returns
    -------
    anim: FuncAnimation
        Composed animation function for all objects / subplots.
    """
    assert len(objects) > 1, 'Cannot compose using a single plot!'
    assert len(set([obj.fig for obj in objects])) == 1, 'All plots must belong to the same figure!'

    fig = objects[0].fig
    if epoch_end == -1:
        epoch_end = min([obj.n_epochs for obj in objects])

    if len(title):
        title += ' - '

    def update(i, objects, epoch_start=0):
        artists = []
        for obj in objects:
            artists += getattr(obj.__class__, '_update')(i, obj, epoch_start)
            for ax, ax_title in zip(obj.axes, obj.title):
                ax.set_title(ax_title)

        obj.fig.suptitle('{}Epoch {}'.format(title, i + epoch_start), fontsize=14)
        obj.fig.tight_layout()
        obj.fig.subplots_adjust(top=0.9)
        return artists

    anim = animation.FuncAnimation(fig, update,
                                   fargs=(objects, epoch_start),
                                   frames=(epoch_end - epoch_start),
                                   blit=True)
    return anim

def compose_plots(objects, epoch, title=''):
    """Compose a single plot from several objects associated with
    subplots of a single figure.

    Parameters
    ----------
    objects: list of plot objects
        Plot objects returned using one of the 'build' methods of the
        Replay class. All the corresponding subplots associated with
        the objects must belong to the same figure.
    epoch: int
        Epoch to use for the plotting.
    title: String, optional
        Text to be used in the title, preceding the epoch information.

    Returns
    -------
    fig: figure
        Figure which contains all subplots.
    """
    assert len(objects) > 1, 'Cannot compose using a single plot!'
    assert len(set([obj.fig for obj in objects])) == 1, 'All plots must belong to the same figure!'

    fig = objects[0].fig
    epoch_end = min([obj.n_epochs for obj in objects])
    epoch = min(epoch, epoch_end)

    for obj in objects:
        getattr(obj.__class__, '_update')(epoch, obj)
        for ax, ax_title in zip(obj.axes, obj.title):
            ax.set_title(ax_title)

    if len(title):
        title += ' - '
    fig.suptitle('{}Epoch {}'.format(title, epoch), fontsize=14)
    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    return fig

class Basic(object):
    """Basic plot class, NOT to be instantiated directly.
    """
    def __init__(self, ax):
        self._title = ''
        self._custom_title = ''
        self.n_epochs = 0

        self.ax = ax
        self.ax.clear()
        self.fig = ax.get_figure()

    @property
    def title(self):
        title = self._title
        if not isinstance(title, tuple):
            title = (self._title,)
        title = tuple([' '.join([self._custom_title, t]) for t in title])
        return title

    @property
    def axes(self):
        return (self.ax,)

    def load_data(self, **kwargs):
        self._prepare_plot()
        return self

    def _prepare_plot(self):
        pass

    @staticmethod
    def _update(i, object, epoch_start=0):
        pass

    def set_title(self, title):
        """Prepends a custom title to the plot.

        Parameters
        ----------
        title: String
            Custom title to prepend.

        Returns
        -------
        None
        """
        self._custom_title = title

    def plot(self, epoch):
        """Plots data at a given epoch.

        Parameters
        ----------
        epoch: int
            Epoch to use for the plotting.

        Returns
        -------
        fig: figure
            Figure containing the plot.
        """
        self.__class__._update(epoch, self)
        self.fig.tight_layout()
        return self.fig

    def animate(self, epoch_start=0, epoch_end=-1):
        """Animates plotted data from `epoch_start` to `epoch_end`.

        Parameters
        ----------
        epoch_start: int, optional
            Epoch to start the animation from.
        epoch_end: int, optional
            Epoch to end the animation.

        Returns
        -------
        anim: FuncAnimation
            Animation function for the data.
        """
        if epoch_end == -1:
            epoch_end = self.n_epochs

        anim = animation.FuncAnimation(self.fig, self.__class__._update,
                                       fargs=(self, epoch_start),
                                       frames=(epoch_end - epoch_start),
                                       blit=True)
        return anim

class FeatureSpace(Basic):
    """Creates an instance of a FeatureSpace object to make plots
    and animations.

    Parameters
    ----------
    ax: AxesSubplot
        Subplot of a Matplotlib figure.
    scaled_fixed: boolean, optional
        If True, axis scales are fixed to the maximum from beginning.
        Default is True.
    """
    def __init__(self, ax, scale_fixed=True):
        super(FeatureSpace, self).__init__(ax)
        self.scale_fixed = scale_fixed
        self.contour = None
        self.bent_inputs = None
        self.bent_lines = None
        self.bent_contour_lines = None
        self.grid_lines = None
        self.contour_lines = None
        self.predictions = None
        self.targets = None

        self.n_inputs = 0

        self.lines = []
        self.points = []

    def load_data(self, feature_space_data):
        """ Loads feature space data as computed in Replay class.

        Parameters
        ----------
        feature_space_data: FeatureSpaceData
            Namedtuple containing information about original grid
            lines, data points and predictions.

        Returns
        -------
        self: FeatureSpace
            Returns the FeatureSpace instance itself.
        """
        self.predictions = feature_space_data.prediction
        self.targets = feature_space_data.target
        self.grid_lines, self.inputs, self.contour_lines = feature_space_data.line
        self.bent_lines, self.bent_inputs, self.bent_contour_lines = feature_space_data.bent_line

        self.n_epochs, _, self.n_inputs = self.bent_inputs.shape

        self.classes = np.unique(self.targets)
        self.bent_inputs = [self.bent_inputs[:, self.targets == target, :] for target in self.classes]

        self._prepare_plot()
        return self

    def _prepare_plot(self):
        if self.scale_fixed:
            xlim = [self.bent_contour_lines[:, :, :, 0].min(), self.bent_contour_lines[:, :, :, 0].max()]
            ylim = [self.bent_contour_lines[:, :, :, 1].min(), self.bent_contour_lines[:, :, :, 1].max()]
            self.ax.set_xlim(xlim)
            self.ax.set_ylim(ylim)

        self.ax.set_xlabel(r"$x_1$", fontsize=14)
        self.ax.set_ylabel(r"$x_2$", fontsize=14, rotation=0)

        self.lines = []
        self.points = []
        for c in range(self.grid_lines.shape[0]):
            line, = self.ax.plot([], [], linewidth=0.5, color='k')
            self.lines.append(line)
        for c in range(len(self.classes)):
            point = self.ax.scatter([], [])
            self.points.append(point)

        contour_x = self.bent_contour_lines[0, :, :, 0]
        contour_y = self.bent_contour_lines[0, :, :, 1]
        self.contour = self.ax.contourf(contour_x, contour_y, np.zeros(shape=(len(contour_x), len(contour_y))),
                              cmap=plt.cm.brg, alpha=0.3, levels=np.linspace(0, 1, 8))

    @staticmethod
    def _update(i, fs, epoch_start=0):
        epoch = i + epoch_start
        fs.ax.set_title('Epoch: {}'.format(epoch))
        if not fs.scale_fixed:
            xlim = [fs.bent_contour_lines[epoch, :, :, 0].min(), fs.bent_contour_lines[epoch, :, :, 0].max()]
            ylim = [fs.bent_contour_lines[epoch, :, :, 1].min(), fs.bent_contour_lines[epoch, :, :, 1].max()]
            fs.ax.set_xlim(xlim)
            fs.ax.set_ylim(ylim)

        if len(fs.lines):
            line_coords = fs.bent_lines[epoch].transpose()

        for c, line in enumerate(fs.lines):
            line.set_data(*line_coords[:, :, c])

        colors = ['b', 'g']
        input_coords = [coord[epoch].transpose() for coord in fs.bent_inputs]
        for c in range(len(fs.points)):
            fs.points[c].remove()
            fs.points[c] = fs.ax.scatter(*input_coords[c], marker='o', color=colors[c], s=10)

        for c in fs.contour.collections:
            c.remove()  # removes only the contours, leaves the rest intact

        fs.contour = fs.ax.contourf(fs.bent_contour_lines[epoch, :, :, 0],
                                    fs.bent_contour_lines[epoch, :, :, 1],
                                    fs.predictions[epoch].squeeze(),
                                    cmap=plt.cm.brg, alpha=0.3, levels=np.linspace(0, 1, 8))

        fs.ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        fs.ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        fs.ax.locator_params(tight=True, nbins=7)
        return fs.lines


class ProbabilityHistogram(Basic):
    """Creates an instance of a ProbabilityHistogram object to make
    plots and animations.

    Parameters
    ----------
    ax1: AxesSubplot
        Subplot of a Matplotlib figure, for the negative cases.
    ax2: AxesSubplot
        Subplot of a Matplotlib figure, for the positive cases.
    """
    def __init__(self, ax1, ax2):
        self._title = ('Negative Cases', 'Positive Cases')
        self._custom_title = ''
        self.ax1 = ax1
        self.ax2 = ax2
        self.ax1.clear()
        self.ax2.clear()
        self.fig = ax1.get_figure()
        self.line = ax1.plot([], [])

        self.proba = None
        self.targets = None
        self.bins = np.linspace(0, 1, 11)

    @property
    def axes(self):
        return (self.ax1, self.ax2)

    def load_data(self, prob_histogram_data):
        """ Loads probability histogram data as computed in Replay
        class.

        Parameters
        ----------
        prob_histogram_data: ProbHistogramData
            Namedtuple containing information about classification
            probabilities and targets.

        Returns
        -------
        self: ProbabilityHistogram
            Returns the ProbabilityHistogram instance itself.
        """
        self.proba, self.targets = prob_histogram_data

        self.n_epochs = self.proba.shape[0]
        self._prepare_plot()
        return self

    def _prepare_plot(self):
        pass

    @staticmethod
    def _update(i, ph, epoch_start=0):
        epoch = i + epoch_start

        correct = ((ph.proba[epoch] > .5) == ph.targets)
        tn = ph.proba[epoch, (ph.targets == 0) & correct]
        fn = ph.proba[epoch, (ph.targets == 0) & ~correct]
        tp = ph.proba[epoch, (ph.targets == 1) & correct]
        fp = ph.proba[epoch, (ph.targets == 1) & ~correct]

        for ax in (ph.ax1, ph.ax2):
            ax.clear()

        ph.ax1.set_title('{} - Epoch: {}'.format(ph.title[0], epoch))
        ph.ax1.set_ylim([0, (ph.targets == 0).sum()])
        ph.ax1.set_xlabel('Probability')
        ph.ax1.set_ylabel('# of Cases')
        ph.ax1.hist(tn, bins=ph.bins, color='k', alpha=.4)
        ph.ax1.hist(fn, bins=ph.bins, color='r', alpha=.5)

        ph.ax2.set_title('{} - Epoch: {}'.format(ph.title[1], epoch))
        ph.ax2.set_ylim([0, (ph.targets == 1).sum()])
        ph.ax2.set_xlabel('Probability')
        ph.ax2.set_ylabel('# of Cases')
        ph.ax2.hist(tp, bins=ph.bins, color='k', alpha=.4)
        ph.ax2.hist(fp, bins=ph.bins, color='r', alpha=.5)

        ph.ax1.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        ph.ax2.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        ph.ax1.locator_params(tight=True, nbins=3)
        ph.ax2.locator_params(tight=True, nbins=3)

        return ph.line

class LossAndMetric(Basic):
    """Creates an instance of a LossAndMetric object to make plots
    and animations.

    Parameters
    ----------
    ax: AxesSubplot
        Subplot of a Matplotlib figure.
    """
    def __init__(self, ax):
        super(LossAndMetric, self).__init__(ax)
        self.ax2 = self.ax.twinx()
        self.line1 = None
        self.line2 = None
        self.point1 = None
        self.point2 = None
        self.metric = None
        self.metric_name = ''

    def load_data(self, loss_and_metric_data):
        """ Loads loss and metric data as computed in Replay class.

        Parameters
        ----------
        loss_and_metric_data: LossAndMetricData
            Namedtuple containing information about loss and a
            given metric.

        Returns
        -------
        self: LossAndMetric
            Returns the LossAndMetric instance itself.
        """
        self.loss, self.metric, self.metric_name = loss_and_metric_data
        self._title = '{} / Loss'.format(self.metric_name)

        self.n_epochs = self.loss.shape[0]
        self._prepare_plot()
        return self

    def _prepare_plot(self):
        self.ax.set_xlim([0, self.n_epochs])
        self.ax.set_xlabel('Epoch')
        self.ax.set_ylim([0, 1.01 * self.metric.max()])
        if self.metric_name == 'acc':
            self.ax.set_ylim([0, 1.01])
        self.ax.set_ylabel(self.metric_name)

        self.ax2.set_xlim([0, self.n_epochs])
        self.ax2.set_ylim([0, 1.01 * self.loss.max()])
        self.ax2.set_ylabel('Loss')

        self.line1, = self.ax.plot([], [], color='k')
        self.line2, = self.ax2.plot([], [], color='r')
        self.point1 = self.ax.scatter([], [], marker='o')
        self.point2 = self.ax2.scatter([], [], marker='o')

        self.ax.legend((self.line1, self.line2), (self.metric_name, 'Loss'), loc=3)

    @staticmethod
    def _update(i, lm, epoch_start=0):
        epoch = i + epoch_start
        lm.ax.set_title('{} - Epoch: {}'.format(lm.title[0], epoch))

        lm.line1.set_data(np.arange(0, epoch + 1), lm.metric[:epoch + 1])
        lm.line2.set_data(np.arange(0, epoch + 1), lm.loss[:epoch + 1])
        lm.point1.remove()
        lm.point1 = lm.ax.scatter(epoch, lm.metric[epoch], marker='o', color='k')
        lm.point2.remove()
        lm.point2 = lm.ax2.scatter(epoch, lm.loss[epoch], marker='o', color='r')

        lm.ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.0f'))
        lm.ax.locator_params(tight=True, nbins=3)

        return lm.line1, lm.line2

class LossHistogram(Basic):
    """Creates an instance of a LossHistogram object to make plots
    and animations.

    Parameters
    ----------
    ax: AxesSubplot
        Subplot of a Matplotlib figure.
    """
    def __init__(self, ax):
        super(LossHistogram, self).__init__(ax)
        self.losses = None
        self._title = 'Losses'

    def __calc_scale(self, margin):
        """ Computes the bins partition for the histogram plot based on the loss range.
        """
        loss_limits = np.array([self.losses.squeeze().min(), self.losses.squeeze().max()])
        loss_range = np.diff(loss_limits)[0]
        exponent = np.floor(np.log10(loss_range))
        magnitude = np.power(10, exponent)
        loss_limits = np.round(loss_limits + np.array([-margin, margin]) * magnitude, exponent.astype(np.int) + 1)
        intervals = (np.diff(loss_limits)[0] / magnitude + 1).astype(np.int)
        while 10 > intervals > 1:
            intervals = (intervals - 1) * 2 + 1
        loss_scale = np.linspace(max(0.0, loss_limits[0]), loss_limits[1], intervals)
        return loss_scale

    def load_data(self, loss_hist_data):
        """ Loads loss histogram data as computed in Replay class.

        Parameters
        ----------
        loss_hist_data: LossHistogramData
            Namedtuple containing information about example's losses.

        Returns
        -------
        self: LossHistogram
            Returns the LossHistogram instance itself.
        """
        self.losses, = loss_hist_data
        self.bins = self.__calc_scale(margin=0)

        self.n_epochs = self.losses.shape[0]
        self._prepare_plot()
        return self

    def _prepare_plot(self):
        self.line = self.ax.plot([], [])

    @staticmethod
    def _update(i, lh, epoch_start=0):
        epoch = i + epoch_start

        lh.ax.clear()

        lh.ax.set_title('{} - Epoch: {}'.format(lh.title[0], epoch))
        lh.ax.set_ylim([0, lh.losses.shape[1]])
        lh.ax.set_xlabel('Loss')
        lh.ax.set_ylabel('# of Cases')
        lh.ax.hist(lh.losses.squeeze()[i], bins=lh.bins, color='k', alpha=.4)

        lh.ax.xaxis.set_major_formatter(ticker.FormatStrFormatter('%0.1f'))
        lh.ax.locator_params(tight=True, nbins=4)

        return lh.line


import numpy as np

def load_data(xlim=(-1, 1), n_points=1000, shuffle=True, seed=13):
    """Generates a dataset composed of two parabolas, with `n_points`
    each. The upper parabola represents the negative cases (class 0)
    and the lower parabola, shifted 0.5 from the first, represents
    the positive cases (class 1).

    Parameters
    ----------
    xlim: tuple of ints, optional
        Boundaries for the X axis. Default is (-1, 1)
    n_points: int, optional
        Number of points in each parabola. Default is 1,000.
    shuffle: boolean, optional
        If True, the points are shuffled. Default is True.
    seed: int, optional
        Random seed. Default is 13.

    Returns
    -------
    X, y: tuple of ndarray
        X is an array of shape (2 * n_points, 2) containing the
        samples for the two parabolas.
        y is an array of shape (2 * n_points, 1) containing the
        classes of the samples.
    """
    # feature x1, 1,000 points evenly spaced between -1 and 1
    x1 = np.linspace(xlim[0], xlim[1], n_points)
    # feature x2, for the two curves
    x2_blue = np.square(x1)
    x2_green = np.square(x1) - .5

    # coordinates for points in the blue line
    blue_line = np.vstack([x1, x2_blue])
    # coordinates for points in the green line
    green_line = np.vstack([x1, x2_green])

    # Remember, blue line is negative (0) and green line is positive (1)
    X = np.concatenate([blue_line, green_line], axis=1).transpose()
    y = np.concatenate([np.zeros(n_points), np.ones(n_points)])

    # But we must not feed the network with neatly organized inputs...
    # so let's randomize them
    if shuffle:
        np.random.seed(seed)
        shuffled = np.random.permutation(range(X.shape[0]))
        X = X[shuffled]
        y = y[shuffled].reshape(-1, 1)

    return (X, y)